static char rcsid[] = "@(#)$Id: sb_file.c,v 2.13 2013-12-01 08:39:36 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.13 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *                  (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "elm_defs.h"
#include "s_me.h"
#include "sb_imp.h"
#include "pg_range_imp.h"

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif

DEBUG_VAR(Debug,__FILE__,"charset");

#if ANSI_C
#define S_(x) static x;
#else
#define S_(x)
#endif

#define SB_in_file_magic     0xEB09

struct sb_in_file {
    unsigned short magic;       /* SB_in_file_magic */

    char * filename;
    FILE * fh;
    int offset_count;
    long * offsets;

    struct sb_pager_range {
	struct pgrng_bfr_binding *pager_range;  /* cache */
	long offset;
    } * pager_range;
    int pager_range_count;

};



static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}

#if 0
static char *us2s P_((unsigned char *str));
static char *us2s(str) 
     unsigned char *str;
{
    return (char *)str;
}
#endif

S_(sb_init_buffer sb_init_file)
static int sb_init_file P_((struct stringbuffer *buffer));
static int sb_init_file(buffer)
     struct stringbuffer *buffer;
{
    static int count =  0;
    int i;
    const char *tmp = give_dt_estr_as_str(&temp_dir_e,"tmpdir",NULL,NULL);

    if (!tmp)
	tmp = "/tmp/";
   

    buffer->a.file = safe_malloc(sizeof (* (buffer->a.file)));
    
    /* defined in hdrs/defs.h */
    bzero((void *)buffer->a.file,sizeof (* (buffer->a.file)));

    buffer->a.file->magic = SB_in_file_magic;

    buffer->a.file->offset_count = 0;
    buffer->a.file->offsets      = NULL;
    buffer->a.file->filename     = NULL;
    buffer->a.file->fh           = NULL;

    buffer->a.file->pager_range = NULL;
    buffer->a.file->pager_range_count = 0;

    for (i = 0; i < 10; i++) {
	count++;

	buffer->a.file->filename = elm_message(FRM("%selmsb-%d-%d"),
						 tmp, getpid (),
						 count);

	buffer->a.file->fh = safeopen_rdwr(buffer->a.file->filename);
	if (!buffer->a.file->fh) {
	    int err UNUSED_VAROK = errno;
	    DPRINT(Debug,25,(&Debug,
			     "stringbuffer: safeopen_rdwr: %s: %s (errno %d)\n",
			     buffer->a.file->filename,
			     strerror(err),err));
	    free(buffer->a.file->filename);
	    buffer->a.file->filename = NULL;
	} else
	    break;
	
    } 
    if (!buffer->a.file->fh)
	return 0;

    unlink (buffer->a.file->filename); 

    DPRINT(Debug,1,(&Debug,
		    "stringbuffer: using temp file (%s)\n",
		    buffer->a.file->filename));
    return 1;
}

S_(sb_free_buffer sb_free_file)
static void sb_free_file P_((struct stringbuffer *buffer));
static void sb_free_file(buffer)
     struct stringbuffer *buffer;
{
    if (SB_in_file_magic != buffer->a.file->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_free_file",
	      "Bad magic number",0);

    if (buffer->a.file->fh) {
	fclose(buffer->a.file->fh);
	buffer->a.file->fh = NULL;
    }

    if (buffer->a.file->filename) {
	free(buffer->a.file->filename);
	buffer->a.file->filename = NULL;
    }

    if (buffer->a.file->offsets) {
	free(buffer->a.file->offsets);
	buffer->a.file->offsets = NULL;
    }
    buffer->a.file->offset_count = 0;

    if (buffer->a.file->pager_range) {
	int i;

	for (i = 0; i < buffer->a.file->pager_range_count; i++) {
	    if (buffer->a.file->pager_range[i].pager_range)
		pgrng_bfr_bnd_clear_stringbuffer(& (buffer->a.file->
						    pager_range[i].pager_range),
						 buffer,i);
	}
    }
    buffer->a.file->pager_range_count = 0;

    buffer->a.file->magic = 0; /* Invalidate */
    free(buffer->a.file);
    buffer->a.file = NULL;
}

static void put_string_serialization_to_file P_((FILE *fh,
						 const struct string *Str));
static void put_string_serialization_to_file(fh,Str)
     FILE *fh; 
     const struct string *Str;
{
    if (RAW_BUFFER == get_string_type(Str)) {
	char * s;
	int l1;

	bytestream_from_string(Str,&s,&l1);
	
	fprintf(fh,"R%d:",l1);
	fputc('\0',fh);  /* Terminating \0 of null charset */

	/* Include ending \0 to string */
	fwrite(s,1,l1+1,fh);
	free(s);

    } else {
	const char * MIME_name = get_string_MIME_name(Str);
	
	if (MIME_name) {
	    char * s;
	    int l1;
	    int l2 = strlen(MIME_name);
	    
	    bytestream_from_string(Str,&s,&l1);
	    
	    fprintf(fh,"%d:%d:",l2,l1);	
	    /* Include ending \0 to string */
	    fwrite(MIME_name,1,l2+1,fh);
	    /* Include ending \0 to string */
	    fwrite(s,1,l1+1,fh);
	    free(s);
	} else { 
	    /* No charset name available ... */
	    struct string *str = convert_string(system_charset,Str,0);
	    char * s;
	    int l1;
	    
	    bytestream_from_string(str,&s,&l1);
	    
	    fprintf(fh,":%d:",l1);
	    fputc('\0',fh);  /* Terminating \0 of null charset */
	    /* Include ending \0 to string */
	    fwrite(s,1,l1+1,fh);
	    free(s);
	    free_string(&str);
	}
    }
}


S_(sb_add_line_to_stringbuffer sb_add_line_to_file)
static void sb_add_line_to_file P_((struct stringbuffer *buffer,
				    const struct string *string,
				    int pg_flags,
				    struct pager_range *range,
				    struct pgrng_bfr_binding  * binding));
static void sb_add_line_to_file(buffer,string,pg_flags,range,binding)
     struct stringbuffer *buffer;
     const struct string *string;
     int pg_flags;
     struct pager_range *range;               /* NOT USED */
     struct pgrng_bfr_binding  * binding;
{
    int ptr;

    if (SB_in_file_magic != buffer->a.file->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_add_line_to_file",
	      "Bad magic number",0);

    ptr = buffer->a.file->offset_count;

    if (fseek(buffer->a.file->fh,0,SEEK_END) != 0) {
	int err UNUSED_VAROK = errno;
	DPRINT(Debug,1,(&Debug,
			"stringbuffer: fseek (flush) failure: %s: %s (errno %d)\n",
			buffer->a.file->filename,
			strerror(err),err));
	
	return;
    }
    
    buffer->a.file->offsets = 
	safe_realloc(buffer->a.file->offsets,
		     (buffer->a.file->offset_count +1) *
		     sizeof (long));
    buffer->a.file->offsets[ptr] = ftell(buffer->a.file->fh);
    buffer->a.file->offset_count++;
    
    if (binding) {
	int range_index;
	struct stringbuffer *bf = 
	    pgrng_bfr_bnd_get_stringbuffer(binding,&range_index);

	if (bf != buffer)
	    panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_add_line_to_file",
		  "pager_range (binding) added to wrong stringbuffer",0);

	if (range_index < 0 || 
	    range_index >=  buffer->a.file->pager_range_count ||
	    buffer->a.file->pager_range[range_index].pager_range != binding)
	    panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_add_line_to_file",
		  "bad range_index",0);
	    
	fprintf(buffer->a.file->fh,"r%d,",range_index);

	free_stringbuffer(&bf);
    }

    fprintf(buffer->a.file->fh,"%d",pg_flags);
    /* : starts string */
    fputc(':', buffer->a.file->fh);
    put_string_serialization_to_file(buffer->a.file->fh,string);
}

S_(sb_linecount_stringbuffer sb_linecount_file)
static int sb_linecount_file P_((const struct stringbuffer *ptr));
static int sb_linecount_file(ptr)
     const struct stringbuffer *ptr;
{
    if (SB_in_file_magic != ptr->a.file->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_linecount_file",
	      "Bad magic number",0);

    return ptr->a.file->offset_count;
}

static struct pager_range * sb_get_pager_range_from_file
  P_((struct stringbuffer * buffer,int index));

struct string * get_string_serialization_from_file P_((FILE *fh));
struct string * get_string_serialization_from_file(fh)
     FILE *fh;
{
    struct string *ret = NULL;

    char * bigs = NULL;
    int setlen = 0;
    int strlen = 0;
    charset_t sn = system_charset;
    int ERRORS = 0;

    int c = fgetc(fh);

    if ('R' == c) {
	sn = RAW_BUFFER;
    } else {
	while (c >= '0' && c <= '9') {
	    setlen = setlen * 10 + c - '0';
	    c = fgetc(fh);
	}
	if (':' != c) {
	    DPRINT(Debug,1,(&Debug,
			    "stringbuffer: charset name len do not end with ':'\n"));
	    
	    goto failure;
	}
    }

    c = fgetc(fh);
    while (c >= '0' && c <= '9') {
	strlen = strlen * 10 + c - '0';
	c = fgetc(fh);
    }
    if (':' != c) {
	DPRINT(Debug,1,(&Debug,
			"stringbuffer: string len do not end with ':'\n"));

	goto failure;
    }

    bigs = safe_malloc(setlen+1+strlen+1);
    if (fread(bigs,1,setlen+1+strlen+1,fh) !=
	setlen+1+strlen+1) {

	goto failure;
    }
    if (bigs[setlen] != '\0' ||
	bigs[setlen+1+strlen] != '\0') {

	goto failure;
    }
	    
    if (setlen != 0) {
	sn = MIME_name_to_charset(bigs,0);
	if (!sn) { 
	    DPRINT(Debug,1,(&Debug,
			    "stringbuffer: Charset '%s' is not known (impossible?)\n",
			    bigs));

	    goto failure;
	}
    }

    ret = new_string(sn);
    
    add_streambytes_to_string(ret,strlen,s2us(&(bigs[setlen+1])),&ERRORS);
    
    if (ERRORS > 0) {
	DPRINT(Debug,1,(&Debug,
			"stringbuffer: %d errors when adding string grom file\n",
			ERRORS));
    }

 failure:
    if (bigs)	    
	free(bigs);

    return ret;
}



S_(sb_get_line_from_stringbuffer sb_get_line_from_file)
static struct string *sb_get_line_from_file P_((struct 
					       stringbuffer *buffer,
						int ptr, int *pg_flags,
						struct pager_range **range));
static struct string *sb_get_line_from_file(buffer,ptr,pg_flags,range)
     struct stringbuffer *buffer;
     int ptr;
     int *pg_flags;
     struct pager_range **range;
{
    int flags = 0;
    struct string * res = NULL;
    int c;
    int range_index = 0;
    int have_range = 0;

    if (SB_in_file_magic != buffer->a.file->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_get_line_from_file",
	      "Bad magic number",0);

    if (ptr < 0 || ptr >= buffer->a.file->offset_count)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_get_line_from_file",
	      "Bad index",0);

    if (0 != fseek(buffer->a.file->fh,buffer->a.file->offsets[ptr],
		   SEEK_SET)) {
	int err UNUSED_VAROK = errno;

	DPRINT(Debug,1,(&Debug,
			"stringbuffer: fseek (flush) failure: %s: %s (errno %d)\n",
			buffer->a.file->filename,
			strerror(err),err));

	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_get_line_from_file",
	      "Failed to flush or seek file",0);
    }

    c = fgetc(buffer->a.file->fh);

    if (range && *range)
	free_pager_range(range);

    if (c == 'r') {
	have_range = 1;

	c = fgetc(buffer->a.file->fh);
	while (c >= '0' && c <= '9') {
	    range_index = range_index * 10 + c - '0';
	    c = fgetc(buffer->a.file->fh);
	}

	if (',' != c) {
	    panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_get_line_from_file",
		  "Bad data read from file",0);
	}
	c = fgetc(buffer->a.file->fh);
    }

    while (c >= '0' && c <= '9') {
	flags = flags * 10 + c - '0';
	c = fgetc(buffer->a.file->fh);
    }
    if (':' != c) {
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_get_line_from_file",
	      "Bad data read from file",0);
    }
    *pg_flags = flags;

    res = get_string_serialization_from_file(buffer->a.file->fh);
    if (!res) {
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_get_line_from_file",
	      "Bad data read from file",0);
    }

    /* May read from file */

    if (range && have_range)
	*range = sb_get_pager_range_from_file(buffer,range_index);
    
    return res;
}

S_(sb_add_pager_range_to_stringbuffer sb_add_pager_range_to_file)
static void sb_add_pager_range_to_file P_((struct stringbuffer * buffer,
					   struct pgrng_bfr_binding * binding,
					   struct pager_range * range));
static void sb_add_pager_range_to_file(buffer,binding,range)
     struct stringbuffer * buffer;
     struct pgrng_bfr_binding * binding;
     struct pager_range * range;
{
    int ptr;

    if (SB_in_file_magic != buffer->a.file->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "sb_add_pager_range_to_file",
	      "Bad magic number",0);

    ptr = buffer->a.file->pager_range_count;

    buffer->a.file->pager_range = 
	safe_realloc(buffer->a.file->pager_range,
		     (buffer->a.file->pager_range_count +1) *
		     sizeof (buffer->a.file->pager_range[0]));
    buffer->a.file->pager_range[ptr].offset = 
	ftell(buffer->a.file->fh);
    buffer->a.file->pager_range[ptr].pager_range = binding;
    pgrng_bfr_bnd_set_stringbuffer(buffer->a.file->pager_range[ptr].
				   pager_range,
				   buffer,ptr);
    buffer->a.file->pager_range_count++;

    /* May need update offset, because put_pager_range_serialization_to_file()
       may call this recursively */
    put_pager_range_serialization_to_file(buffer->a.file->fh,range,buffer,
					  & buffer->a.file->pager_range[ptr].offset);
}


S_(sb_get_pager_range_from_stringbuffer sb_get_pager_range_from_file)
static struct pager_range * sb_get_pager_range_from_file(buffer,index)
     struct stringbuffer * buffer;
     int index;
{
    struct pager_range *ret = NULL;

    if (SB_in_file_magic != buffer->a.file->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "sb_get_pager_range_from_file",
	      "Bad magic number",0);

    if (index < 0 || index >= buffer->a.file->pager_range_count)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "sb_get_pager_range_from_file",
	      "Bad index",0);
	
    if (! buffer->a.file->pager_range[index].pager_range) {
	buffer->a.file->pager_range[index].pager_range = new_pgrng_bfr_binding();
	pgrng_bfr_bnd_set_stringbuffer(buffer->a.file->pager_range[index].pager_range,
				       buffer,index);
    }
	
    ret = pgrng_bfr_bnd_get_range(buffer->a.file->pager_range[index].pager_range);

    if (!ret) {

	if (0 != fseek(buffer->a.file->fh,buffer->a.file->pager_range[index].offset,
		       SEEK_SET)) {
	    int err UNUSED_VAROK = errno;

	    DPRINT(Debug,1,(&Debug,
			    "stringbuffer: fseek (flush) failure: %s: %s (errno %d)\n",
			    buffer->a.file->filename,
			    strerror(err),err));
	    panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_get_pager_range_from_file",
		  "Failed to flush or seek file",0);
	}

	ret = get_pager_range_serialization_from_file(buffer->a.file->fh,buffer);

	if (!ret)
	   panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_get_pager_range_from_file",
		 "No pager_range on file",0); 

	set_pager_range_stringbuffer(ret,
				     buffer->a.file->pager_range[index].pager_range);
    }

    return ret;
}

struct sb_type   sb_in_file = {
    SB_TYPE_magic,

    sb_init_file,
    sb_free_file,
    sb_add_line_to_file,
    sb_linecount_file,
    sb_get_line_from_file,
    sb_add_pager_range_to_file,
    sb_get_pager_range_from_file

};

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */

