static char rcsid[] = "@(#)$Id: stringbuffer.c,v 2.8 2013-10-13 07:18:53 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.8 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *                  (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "elm_defs.h"
#include "s_me.h"
#include "sb_imp.h"
#include "pg_range_imp.h"

DEBUG_VAR(Debug,__FILE__,"charset");

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif

static struct stringbuffer *malloc_buffer P_((struct sb_type *tag));
static struct stringbuffer *malloc_buffer(tag)
     struct sb_type *tag;
{
    struct stringbuffer *ret = safe_malloc(sizeof (*ret));

    if (SB_TYPE_magic != tag->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"malloc_buffer",
	      "Bad magic number (buffer type)",0);

    /* defined in hdrs/defs.h */
    bzero((void *)ret,sizeof (struct stringbuffer));

    ret->magic               = STRINGBUFFER_magic;

    ret->refcount            = 1;

    ret->buffer_type         = tag;
    ret->a.dummy             = NULL;

    if (!ret->buffer_type->sb_init_it(ret)) {
	DPRINT(Debug,68,(&Debug,
			 "-- failed to init membuffer (type=%p)\n",
			 tag));
	ret->buffer_type->sb_free_it(ret);
	ret->buffer_type = NULL;
	ret->magic       = 0; /* Invvalidate */
	free(ret);
	return NULL;
    }

    return ret;
}

struct stringbuffer * create_membuffer()
{
    struct stringbuffer *ret = malloc_buffer(&sb_in_mem);
    
    if (!ret) {
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"create_membuffer",
	      "Initialization failure",0);	
    }

    DPRINT(Debug,68,(&Debug,
		     "create_membuffer=%p (type=%p)\n",
		     ret,ret->buffer_type));

    return ret;
}

struct stringbuffer * create_filebuffer()
{
    struct stringbuffer *ret = malloc_buffer(&sb_in_file);
    
    if (!ret) {
	int err = errno;
	lib_error(CATGETS(elm_msg_cat, MeSet, MeCantCreaTempFIleUsingMemory,
			  "Sorry, can't create temp file (error: %s) using memory"),
		  strerror(err));
	ret = malloc_buffer(&sb_in_mem);
	if (!ret) 
	    panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"create_membuffer",
		  "Initialization failure",0);	
    }

    DPRINT(Debug,68,(&Debug,
		     "create_membuffer=%p (type=%p)\n",
		     ret,ret->buffer_type));

    return ret;
}



void free_stringbuffer(ptr)
     struct stringbuffer **ptr;
{
    if (STRINGBUFFER_magic != (*ptr)->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"free_stringbuffer",
	      "Bad magic number (buffer type)",0);

    if (SB_TYPE_magic != (*ptr)->buffer_type->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"free_stringbuffer",
	      "Bad magic number (buffer type)",0);

    DPRINT(Debug,68,(&Debug,
		     "free_stringbuffer(*ptr=%p) (type=%p)\n",
		     (*ptr),(*ptr)->buffer_type));

    (*ptr)->refcount--;

    if ((*ptr)->refcount > 0) {  /* Do not free */
	
        (*ptr) = NULL;           /* Refefence count for this pointer is     */
	return;			 /* decrement, so this must have be reset   */
	                                 
    }
	

    (*ptr)->buffer_type->sb_free_it(*ptr);

    (*ptr)->buffer_type = NULL;
    (*ptr)->magic       = 0;     /* Invalidate */
    free(*ptr);
    *ptr = NULL;
}

void inc_stringbuffer_refcount(buffer)
     struct stringbuffer *buffer;
{
    if (STRINGBUFFER_magic != buffer->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "inc_stringbuffer_refcount",
	      "Bad magic number (buffer type)",0);

    buffer->refcount++;
}

void add_line_to_stringbuffer(buffer,string,pg_flags,range)
     struct stringbuffer *buffer;
     const struct string *string;
     int pg_flags;
     struct pager_range *range; 
{
    const char * MIME_name                    UNUSED_VAROK =  
	get_string_MIME_name(string);
    int len                                   UNUSED_VAROK = 
	string_len(string);
    const struct  charset_type * charset_type UNUSED_VAROK =
	get_string_charset_type(string);
    struct pgrng_bfr_binding  * binding = NULL;

    if (STRINGBUFFER_magic != buffer->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "add_line_to_stringbuffer",
	      "Bad magic number (buffer type)",0);

    if (SB_TYPE_magic != buffer->buffer_type->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "add_line_to_stringbuffer",
	      "Bad magic number (buffer type)",0);

    DPRINT(Debug,68,(&Debug,
		     "add_line_to_stringbuffer(%p,%p,pg_flags=%d) (type=%p, string type=('%s'; type=%p)) -- string len = %d\n",
		     buffer,string,
		     pg_flags,
		     buffer->buffer_type,
		     MIME_name ? MIME_name : "<none>",
		     charset_type,len));
#if 0
    DPRINT(Debug,69,(&Debug," .... string=%S\n",string));
#endif

    if (range) {
	binding = give_pager_range_stringbuffer(range);

	if (!binding)
	    panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
		  "add_line_to_stringbuffer",
		  "pager_range not added to stringbuffer",0);

    }


    buffer->buffer_type->sb_add_line_to_it(buffer,string,pg_flags,range,binding);

    if (binding)
	free_pgrng_bfr_binding(&binding);
}

int linecount_stringbuffer(ptr)
     const struct stringbuffer *ptr;
{
    int ret;


    if (STRINGBUFFER_magic != ptr->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "linecount_stringbuffer",
	      "Bad magic number (buffer type)",0);

    if (SB_TYPE_magic != ptr->buffer_type->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"linecount_stringbuffer",
	      "Bad magic number (buffer type)",0);

    ret=ptr->buffer_type->sb_linecount_it(ptr);

    DPRINT(Debug,68,(&Debug,
		     "linecount_stringbuffer(%p)=%d (type=%p)\n",
		     ptr,ret,ptr->buffer_type));

    return ret;
}

struct string *get_line_from_stringbuffer(buffer,ptr,pg_flags,
					  range_changed,range)
     struct stringbuffer *buffer;
     int ptr;
     int *pg_flags;
     int *range_changed;
     struct pager_range **range;  
{
    struct string *ret;
    int PG_FLAGS = 0;
    struct pager_range * new_range = NULL;

    if (STRINGBUFFER_magic != buffer->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "get_line_from_stringbuffer",
	      "Bad magic number (buffer type)",0);

    if (SB_TYPE_magic != buffer->buffer_type->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "get_line_from_stringbuffer",
	      "Bad magic number (buffer type)",0);

    if (range_changed)
	*range_changed = 0;


    ret = buffer->buffer_type->sb_get_line_from_it(buffer,ptr,&PG_FLAGS,
						   range ? &new_range : NULL);

    if (range && *range != new_range) {
	if (range_changed)
	    *range_changed = 1;
	
	if (*range)
	    free_pager_range(range);
	
	*range = new_range;
	new_range = NULL;
    }

    if (new_range)
	free_pager_range(&new_range);


    {
	const char * MIME_name                    UNUSED_VAROK =  
	    get_string_MIME_name(ret);
	const struct  charset_type * charset_type UNUSED_VAROK =
	    get_string_charset_type(ret);

	DPRINT(Debug,68,(&Debug,
			 "get_line_from_stringbuffer(%p,%d)=%p (type=%p; string type=('%s'; type=%p), pg_flags=%d)\n",
			 buffer,ptr,ret,
			 buffer->buffer_type,
			 MIME_name ? 
			 MIME_name : "<none>",
			 charset_type,
			 PG_FLAGS));
    }

    if (pg_flags)
	*pg_flags = PG_FLAGS;

#if 0
    DPRINT(Debug,69,(&Debug," .... return=%S\n",ret));
#endif

    return ret;
}

/* Allocates new struct pager_range -- not used */
struct pager_range * stringbuffer_add_simple_pager_range(buffer,inherit,
							 range_flags,
							 quote_level)
     struct stringbuffer *buffer;
     struct pager_range *inherit;
     int range_flags;
     int quote_level;
{
    struct pager_range * ret = NULL;

    if (STRINGBUFFER_magic != buffer->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "stringbuffer_add_pager_range",
	      "Bad magic number (buffer type)",0);

    ret = malloc_simple_pager_range(inherit,range_flags,quote_level);

    stringbuffer_add_pager_range(buffer,ret);

    return ret;
}

void stringbuffer_add_pager_range(buffer,range)
     struct stringbuffer *buffer;
     struct pager_range *range;
{
    struct pgrng_bfr_binding *binding = new_pgrng_bfr_binding();
    struct pager_range * inherited, *child;
   
    if (STRINGBUFFER_magic != buffer->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "stringbuffer_add_pager_range",
	      "Bad magic number (buffer type)",0);

    set_pager_range_stringbuffer(range,binding);


    child = range;
    inc_pager_range_refcount(child);

    /* Check parents */
    while (NULL != (inherited = give_parent_pager_range(child))) {
	struct pgrng_bfr_binding * bng_inherited = give_pager_range_stringbuffer(inherited);
	struct stringbuffer  * buf_inherited;

	if (! bng_inherited) 
	    stringbuffer_add_pager_range(buffer,inherited);
	if (! bng_inherited) 
	    bng_inherited = give_pager_range_stringbuffer(inherited);
	
	buf_inherited = pgrng_bfr_bnd_get_stringbuffer(bng_inherited,NULL);

	if (! buf_inherited)
	    stringbuffer_add_pager_range_b(buffer,bng_inherited,inherited);

	if (! buf_inherited)
	    buf_inherited = pgrng_bfr_bnd_get_stringbuffer(bng_inherited,NULL);

	if (buf_inherited  != buffer)
	    if (STRINGBUFFER_magic != buffer->magic)
		panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
		      "stringbuffer_add_pager_range",
		      "Parent/inherited pager_range added to othher stringbuffer",0);
	
	free_stringbuffer(&buf_inherited);

	free_pgrng_bfr_binding(&bng_inherited);
	free_pager_range(&child);
	child = inherited;
    }
    free_pager_range(&child);

    stringbuffer_add_pager_range_b(buffer,binding,range);
 
    free_pgrng_bfr_binding(&binding);
}

void stringbuffer_add_pager_range_b(buffer,binding,range)
     struct stringbuffer *buffer;
     struct pgrng_bfr_binding *binding;
     struct pager_range *range;
{
    if (STRINGBUFFER_magic != buffer->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "stringbuffer_add_pager_range_b",
	      "Bad magic number (buffer type)",0);

    if (SB_TYPE_magic != buffer->buffer_type->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "stringbuffer_add_simple_pager_range_b",
	      "Bad magic number (buffer type)",0);
    
    buffer->buffer_type->sb_add_pager_range_to_it(buffer,binding,range);

}


/* Caller must call free_pager_range() 
   used by get_pager_range_serialization_from_file()
*/
struct pager_range * stringbuffer_get_pager_range(buffer,index)
     struct stringbuffer *buffer; 
     int index; 
{
    if (STRINGBUFFER_magic != buffer->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "stringbuffer_get_pager_range",
	      "Bad magic number (buffer type)",0);

    if (SB_TYPE_magic != buffer->buffer_type->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "stringbuffer_get_pager_range",
	      "Bad magic number (buffer type)",0);

    return buffer->buffer_type->sb_get_pager_range_from_it(buffer,index);

}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */

